import { FastifyInstance, RawServerBase, RawServerDefault, RawRequestDefaultExpression, RawReplyDefaultExpression, ContextConfigDefault, FastifyRequest, FastifyReply, FastifyLoggerInstance } from 'fastify';
import { RouteGenericInterface } from 'fastify/types/route';
import { MongoClient, Collection } from 'mongodb';
export interface FnfFastifyRequest<RouteGeneric extends RouteGenericInterface = RouteGenericInterface, RawServer extends RawServerBase = RawServerDefault, RawRequest extends RawRequestDefaultExpression<RawServer> = RawRequestDefaultExpression<RawServer>> extends FastifyRequest {
    authorization: boolean;
    authenticated: boolean;
    type: 'getOne' | 'getAll' | 'post' | 'patch' | 'delete';
    document?: unknown;
    documents?: unknown;
    collection?: Collection;
    total?: number;
    filterQuery?: {
        [key: string]: any;
    };
}
export interface FnfFastifyInstance<RawServer extends RawServerBase = RawServerDefault, RawRequest extends RawRequestDefaultExpression<RawServer> = RawRequestDefaultExpression<RawServer>, RawReply extends RawReplyDefaultExpression<RawServer> = RawReplyDefaultExpression<RawServer>, Logger = FastifyLoggerInstance> extends FastifyInstance {
    mongo: MongoClient;
}
export interface FnfRouteHandlerMethod<RawServer extends RawServerBase = RawServerDefault, RawRequest extends RawRequestDefaultExpression<RawServer> = RawRequestDefaultExpression<RawServer>, RawReply extends RawReplyDefaultExpression<RawServer> = RawReplyDefaultExpression<RawServer>, RouteGeneric extends RouteGenericInterface = RouteGenericInterface, ContextConfig = ContextConfigDefault> extends Function {
    (this: FnfFastifyInstance<RawServer, RawRequest, RawReply>, request: FnfFastifyRequest<RouteGeneric, RawServer, RawRequest>, reply: FastifyReply<RawServer, RawRequest, RawReply, RouteGeneric, ContextConfig>, next?: () => void): void | Promise<any>;
    getOne?: boolean;
    getAll?: boolean;
    post?: boolean;
    patch?: boolean;
    delete?: boolean;
}
export interface Route {
    uri: string;
    prefix: string;
    type: string;
}
export interface RouteModule {
    attributes: Object;
    requiredAttributes: string[];
    requiredPatchAttributes: string[];
    relationships: Object;
    requiredRelationships: string[];
    requiredPatchRelationships: string[];
    route: Route;
    type: string;
    authorization: boolean | {
        getOne: boolean;
        getAll: boolean;
        post: boolean;
        patch: boolean;
        delete: boolean;
    };
    preHooks?: FnfRouteHandlerMethod[];
    postHooks?: FnfRouteHandlerMethod[];
    getOne?: FnfRouteHandlerMethod;
    getAll?: FnfRouteHandlerMethod;
    post?: FnfRouteHandlerMethod;
    patch?: FnfRouteHandlerMethod;
    delete?: FnfRouteHandlerMethod;
    perPage?: number;
}
export interface ServerOptions {
    port?: number;
    preHooks?: FnfRouteHandlerMethod[];
    postHooks?: FnfRouteHandlerMethod[];
    authentication?: FnfRouteHandlerMethod;
}
export interface GetQueryParams {
    include?: string;
}
export interface GetAllQueryParams {
    page?: string;
    filter?: string;
    orderBy?: string;
    order?: 'ascending' | 'descending';
    include?: string;
}
export interface GetParams {
    id: string;
}
export interface PatchParams {
    id: string;
}
export interface DeleteParams {
    id: string;
}
export default function (callback: (server: FnfFastifyInstance) => Promise<void>, opts?: ServerOptions): Promise<void>;
export declare class FnfDone {
}
export declare class FnfError {
    options: JsonApiErrorOptions;
    constructor(options: JsonApiErrorOptions);
}
export declare const jsonapiResponse: (data: Object | Object[], included?: any[]) => Promise<any>;
export interface JsonApiErrorOptions {
    request: FnfFastifyRequest;
    reply: FastifyReply;
    status: number;
    title: string;
    detail?: string;
    code?: string;
}
export declare const jsonapiError: ({ request, reply, status, title, detail, code, }: JsonApiErrorOptions) => Promise<{
    errors: {
        id: string;
        status: number;
        code: string;
        title: string;
        detail: string;
    }[];
}>;
