"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.jsonapiError = exports.jsonapiResponse = exports.FnfError = exports.FnfDone = void 0;
require('source-map-support').install();
const fs = __importStar(require("fs"));
const path = __importStar(require("path"));
const fastify_1 = __importDefault(require("fastify"));
const mongodb_1 = require("mongodb");
const uuid_1 = require("uuid");
const boolean_1 = require("boolean");
const get_1 = __importDefault(require("lodash/get"));
const qs_1 = __importDefault(require("qs"));
const PREFIX = process.env.PREFIX || '';
const ROUTE_DIR = path.join('dist', 'routes');
const DB_NAME = process.env.MONGO_DATABASE || 'test';
const MONGO_CREDENTIALS = process.env.MONGO_USERNAME ? `${process.env.MONGO_USERNAME}:${process.env.MONGO_PASSWORD}@` : '';
const MONGO_HOST = process.env.MONGO_HOST || 'localhost';
const MONGO_VALIDATE_SSL = boolean_1.boolean(process.env.MONGO_VALIDATE_SSL || true);
// const MODULE_NAME = process.env.npm_package_name;
const DEFAULT_PER_PAGE = 10;
let server;
const defaultOptions = {
    port: 3123,
};
let options;
function default_1(callback, opts = {}) {
    return __awaiter(this, void 0, void 0, function* () {
        options = opts;
        Object.keys(defaultOptions).forEach((key) => {
            if (typeof options[key] === 'undefined') {
                options[key] = defaultOptions[key];
            }
        });
        // Instantiate mongodb
        // const uri = "mongodb+srv://<username>:<password>@<your-cluster-url>/test?retryWrites=true&w=majority";
        const uri = `mongodb://${MONGO_CREDENTIALS}${MONGO_HOST}:27017`;
        const client = new mongodb_1.MongoClient(uri, { useUnifiedTopology: true, useNewUrlParser: true, ssl: true, sslValidate: MONGO_VALIDATE_SSL });
        const mongo = yield client.connect();
        // Instantiate fastify
        server = fastify_1.default({
            logger: true,
        });
        server.decorate('mongo', mongo);
        yield loadRoutes(server);
        // Declare a route
        server.get('/', (request, reply) => __awaiter(this, void 0, void 0, function* () {
            reply.send({ hello: 'world' });
        }));
        yield callback(server);
        // Run the server!
        server.listen(options.port, '0.0.0.0', () => ((err) => __awaiter(this, void 0, void 0, function* () {
            if (err) {
                server.log.error(err);
                process.exit(1);
            }
        }))().catch(console.error));
    });
}
exports.default = default_1;
const loadRoutes = (server) => __awaiter(void 0, void 0, void 0, function* () {
    const routes = new Set();
    const loadFiles = (prefix) => __awaiter(void 0, void 0, void 0, function* () {
        let routeFiles;
        try {
            routeFiles = yield fs.promises.readdir(prefix);
        }
        catch (e) {
            server.log.error('Error accessing routes');
            throw e;
        }
        for (const filename of routeFiles) {
            try {
                const filePath = path.join(prefix, filename);
                const stats = yield fs.promises.stat(filePath);
                if (stats.isDirectory()) {
                    routes.add({
                        uri: filePath.replace(ROUTE_DIR, ''),
                        prefix,
                        type: filename,
                    });
                    yield loadFiles(filePath);
                }
            }
            catch (e) { }
        }
    });
    yield loadFiles(ROUTE_DIR);
    for (const route of routes) {
        const routeModule = require(fs.realpathSync(`./${path.join(ROUTE_DIR, route.uri)}`));
        server.log.info({ route, attributes: routeModule.attributes });
        routeModule.route = route;
        if (!routeModule.type) {
            routeModule.type = route.type;
        }
        let relationships;
        if (routeModule.relationships) {
            relationships = {};
            for (const relationshipName of Object.keys(routeModule.relationships)) {
                const relationship = routeModule.relationships[relationshipName];
                let typeName = relationshipName;
                if (relationship.relationship) {
                    typeName = relationship.relationship;
                }
                if (relationship.many) {
                    relationships[relationshipName] = {
                        type: 'array',
                        items: {
                            type: 'object',
                            properties: {
                                type: {
                                    type: 'string',
                                    enum: [typeName],
                                },
                                id: {
                                    type: 'string',
                                },
                            },
                            required: ['type', 'id'],
                        },
                    };
                }
                else {
                    relationships[relationshipName] = {
                        type: 'object',
                        properties: {
                            type: {
                                type: 'string',
                                enum: [typeName],
                            },
                            id: {
                                type: 'string',
                            },
                        },
                        required: ['type', 'id'],
                    };
                }
            }
        }
        const validResponseObject = {
            type: 'object',
            properties: {
                id: {
                    type: 'string',
                },
                type: {
                    type: 'string',
                },
                attributes: {
                    type: 'object',
                    properties: Object.assign({ createdAt: {
                            type: ['string', 'null'],
                            format: 'date-time',
                        }, updatedAt: {
                            type: ['string', 'null'],
                            format: 'date-time',
                        }, deletedAt: {
                            type: ['string', 'null'],
                            format: 'date-time',
                        } }, routeModule.attributes),
                },
                meta: {
                    type: 'object',
                }
            },
        };
        const validListResponse = {
            type: 'object',
            properties: {
                data: {
                    type: 'array',
                    items: validResponseObject,
                },
            },
        };
        const validSingleResponse = {
            type: 'object',
            properties: {
                data: validResponseObject,
            },
        };
        const validPost = {
            type: 'object',
            properties: {
                data: {
                    type: 'object',
                    properties: {
                        type: {
                            type: 'string',
                            enum: [routeModule.type]
                        },
                        attributes: {
                            type: 'object',
                            properties: routeModule.attributes,
                            required: routeModule.requiredAttributes || Object.keys(routeModule.attributes),
                        },
                    },
                    required: [
                        'type',
                        'attributes',
                    ],
                },
            },
            required: [
                'data',
            ],
        };
        const validPatch = {
            type: 'object',
            properties: {
                data: {
                    type: 'object',
                    properties: {
                        type: {
                            type: 'string',
                            enum: [routeModule.type]
                        },
                        attributes: {
                            type: 'object',
                            properties: routeModule.attributes,
                            required: routeModule.requiredPatchAttributes || [],
                        },
                    },
                    required: [
                        'type',
                        'attributes',
                    ],
                },
            },
            required: [
                'data',
            ],
        };
        if (relationships) {
            validPost.properties.data.properties['relationships'] = {
                type: 'object',
                properties: relationships,
                required: routeModule.requiredRelationships || Object.keys(relationships),
            };
            if (routeModule.requiredRelationships && routeModule.requiredRelationships.length) {
                validPost.properties.data.required.push('relationships');
            }
            validPatch.properties.data.properties['relationships'] = {
                type: 'object',
                properties: relationships,
                required: routeModule.requiredPatchRelationships || [],
            };
            if (routeModule.requiredPatchRelationships && routeModule.requiredPatchRelationships.length) {
                validPatch.properties.data.required.push('relationships');
            }
            const getSchema = {
                type: 'object',
                properties: relationships,
            };
            validSingleResponse.properties.data.properties['relationships'] = getSchema;
            validListResponse.properties.data.items.properties['relationships'] = getSchema;
        }
        server.get(`${PREFIX}${route.uri}/:id`, {
        // schema: {
        //   response: {
        //     200: validSingleResponse,
        //   },
        // },
        }, yield defaultRouteGetOne(routeModule));
        server.get(`${PREFIX}${route.uri}`, {
            schema: {
                response: {
                // 200: validListResponse,
                },
            },
        }, yield defaultRouteGetAll(routeModule));
        server.post(`${PREFIX}${route.uri}`, {
            schema: {
                body: validPost,
                response: {
                    200: validSingleResponse,
                },
            },
        }, yield defaultRoutePost(routeModule));
        server.patch(`${PREFIX}${route.uri}/:id`, {
            schema: {
                body: validPatch,
                response: {
                    200: validSingleResponse,
                },
            },
        }, yield defaultRoutePatch(routeModule));
        server.delete(`${PREFIX}${route.uri}/:id`, {
            schema: {
                response: {
                    204: {
                        type: 'null',
                    },
                },
            },
        }, yield defaultRouteDelete(routeModule));
    }
});
class FnfDone {
}
exports.FnfDone = FnfDone;
class FnfError {
    constructor(options) {
        this.options = options;
    }
}
exports.FnfError = FnfError;
const setRequestType = (request) => __awaiter(void 0, void 0, void 0, function* () {
    if (request.method === 'GET') {
        if (request.params.id) {
            request.type = 'getOne';
        }
        else {
            request.type = 'getAll';
        }
    }
    else if (request.method === 'POST') {
        request.type = 'post';
    }
    else if (request.method === 'PATCH') {
        request.type = 'patch';
    }
    else if (request.method === 'DELETE') {
        request.type = 'delete';
    }
});
const doAuth = (request, reply, routeModule) => __awaiter(void 0, void 0, void 0, function* () {
    if (typeof routeModule.authorization === 'undefined')
        return;
    if (routeModule.authorization === true || routeModule.authorization[request.type]) {
        request.authorization = true;
        if (options.authentication) {
            yield options.authentication.bind(server)(request, reply);
            if (!request.authenticated) {
                throw new FnfError({
                    request,
                    reply,
                    status: 401,
                    title: 'Unauthorized',
                    detail: 'This request requires authentication.',
                    code: 'UNAUTHORIZED',
                });
            }
        }
    }
});
const doPreHooks = (request, reply, routeModule) => __awaiter(void 0, void 0, void 0, function* () {
    let nextCalled = true;
    const next = () => {
        nextCalled = true;
    };
    if (options.preHooks) {
        nextCalled = false;
        for (const middleFunc of options.preHooks) {
            if (middleFunc[request.type]) {
                yield middleFunc.bind(server)(request, reply, next);
                if (nextCalled) {
                    nextCalled = false;
                }
                else {
                    break;
                }
            }
        }
    }
    if (routeModule.preHooks) {
        if (!nextCalled)
            return;
        nextCalled = false;
        for (const middleFunc of routeModule.preHooks) {
            if (middleFunc[request.type]) {
                yield middleFunc.bind(server)(request, reply, next);
                if (nextCalled) {
                    nextCalled = false;
                }
                else {
                    break;
                }
            }
        }
    }
    return nextCalled;
});
const preHooks = (request, reply, routeModule) => __awaiter(void 0, void 0, void 0, function* () {
    yield setRequestType(request);
    yield doAuth(request, reply, routeModule);
    return doPreHooks(request, reply, routeModule);
});
const doPostHooks = (request, reply, routeModule) => __awaiter(void 0, void 0, void 0, function* () {
    if (options.postHooks) {
        for (const middleFunc of options.postHooks) {
            if (middleFunc[request.type])
                yield middleFunc.bind(server)(request, reply);
        }
    }
    if (routeModule.postHooks) {
        for (const middleFunc of routeModule.postHooks) {
            if (middleFunc[request.type])
                yield middleFunc.bind(server)(request, reply);
        }
    }
});
const postHooks = (request, reply, routeModule) => __awaiter(void 0, void 0, void 0, function* () {
    yield doPostHooks(request, reply, routeModule);
});
const handleError = (request, reply, routeModule, e) => __awaiter(void 0, void 0, void 0, function* () {
    if (e instanceof FnfDone) {
        // intentionally blank
    }
    else if (e instanceof FnfError) {
        const status = e.options.status;
        reply.status(status).send(yield exports.jsonapiError(e.options));
    }
    else if (e instanceof Error) {
        server.log.error(e);
        console.error(e);
        const status = 500;
        reply.status(status).send(yield exports.jsonapiError({
            request,
            reply,
            status,
            title: 'Internal Server Error',
            code: 'UNKNOWN_ERROR',
            detail: e.message,
        }));
    }
    else {
        server.log.error(e);
        const status = 500;
        reply.status(status).send(yield exports.jsonapiError({
            request,
            reply,
            status,
            title: 'Internal Server Error',
            code: 'UNKNOWN_ERROR',
        }));
    }
});
const fetchIncludes = function (document, include, included) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!included)
            included = [];
        if (Array.isArray(document)) {
            yield document.reduce((promise, doc) => __awaiter(this, void 0, void 0, function* () {
                yield promise;
                included = yield fetchIncludes.bind(this)(doc, include, included);
            }), Promise.resolve());
            return included;
        }
        if (!document.relationships)
            return;
        const db = this.mongo.db(DB_NAME);
        const includes = include.split(',');
        includes.sort();
        if (includes.length > 1) {
            return includes.reduce((promise, include) => __awaiter(this, void 0, void 0, function* () {
                yield promise;
                return fetchIncludes.bind(this)(document, include, included);
            }), Promise.resolve());
        }
        if (include === '*') {
            for (const key of Object.keys(document.relationships)) {
                if (document.relationships[key] && Array.isArray(document.relationships[key].data)) {
                    if (document.relationships[key].data.length === 0)
                        continue;
                    const collection = db.collection(document.relationships[key].data[0].type);
                    for (const relationship of document.relationships[key].data) {
                        if (included.some((doc) => doc.id === relationship.id && doc.type === relationship.type))
                            continue;
                        const includedDoc = yield collection.findOne({
                            'attributes.deletedAt': {
                                $exists: false,
                            },
                            id: {
                                $eq: relationship.id,
                            },
                        }, {
                            projection: {
                                _id: 0,
                            },
                        });
                        if (includedDoc) {
                            included.push(includedDoc);
                        }
                    }
                }
                else if (document.relationships[key] && document.relationships[key].data) {
                    const collection = db.collection(document.relationships[key].data.type);
                    const includedDoc = yield collection.findOne({
                        'attributes.deletedAt': {
                            $exists: false,
                        },
                        id: {
                            $eq: document.relationships[key].data.id,
                        },
                    }, {
                        projection: {
                            _id: 0,
                        },
                    });
                    if (includedDoc) {
                        included.push(includedDoc);
                    }
                }
            }
        }
        else if (include === '*.*') {
            yield included.reduce((promise, doc) => __awaiter(this, void 0, void 0, function* () {
                yield promise;
                yield fetchIncludes.bind(this)(doc, '*', included);
            }), Promise.resolve());
        }
        else {
            if (document.relationships[include] && Array.isArray(document.relationships[include].data) && document.relationships[include].data.length > 0) {
                const collection = db.collection(document.relationships[include].data[0].type);
                for (const relationship of document.relationships[include].data) {
                    if (included.some((doc) => doc.id === relationship.id && doc.type === relationship.type))
                        continue;
                    const includedDoc = yield collection.findOne({
                        'attributes.deletedAt': {
                            $exists: false,
                        },
                        id: {
                            $eq: relationship.id,
                        },
                    }, {
                        projection: {
                            _id: 0,
                        },
                    });
                    if (includedDoc) {
                        included.push(includedDoc);
                    }
                }
            }
            else if (document.relationships[include] && !Array.isArray(document.relationships[include].data) && document.relationships[include].data) {
                const collection = db.collection(document.relationships[include].data.type);
                const includedDoc = yield collection.findOne({
                    'attributes.deletedAt': {
                        $exists: false,
                    },
                    id: {
                        $eq: document.relationships[include].data.id,
                    },
                }, {
                    projection: {
                        _id: 0,
                    },
                });
                if (includedDoc) {
                    included.push(includedDoc);
                }
            }
        }
        return included;
    });
};
const defaultRouteGetOne = (routeModule) => __awaiter(void 0, void 0, void 0, function* () {
    return (request, reply) => __awaiter(void 0, void 0, void 0, function* () {
        const params = request.params;
        const query = qs_1.default.parse(qs_1.default.stringify(request.query));
        try {
            const db = server.mongo.db(DB_NAME);
            request.collection = db.collection(routeModule.route.type);
            request.document = yield request.collection.findOne({
                'attributes.deletedAt': {
                    $exists: false,
                },
                id: {
                    $eq: params.id,
                },
            });
            if (request.document)
                delete request.document['_id'];
            // if middleware did not call next, do not perform default actions
            let nextCalled = yield preHooks(request, reply, routeModule);
            if (!nextCalled)
                return;
            if (routeModule.getOne) {
                nextCalled = false;
                yield routeModule.getOne.bind(server)(request, reply, () => {
                    nextCalled = true;
                });
            }
            if (!nextCalled)
                return;
            if (!request.document) {
                throw new FnfError({
                    request,
                    reply,
                    status: 404,
                    title: 'Not Found',
                    code: 'NOT_FOUND',
                    detail: 'The requested resource was not found.',
                });
            }
            const included = query.include ? yield fetchIncludes.bind(server)(request.document, query.include) : undefined;
            reply.status(200).send(yield exports.jsonapiResponse(request.document, included));
        }
        catch (e) {
            yield handleError(request, reply, routeModule, e);
        }
    });
});
const defaultRouteGetAll = (routeModule) => __awaiter(void 0, void 0, void 0, function* () {
    return (request, reply) => __awaiter(void 0, void 0, void 0, function* () {
        const query = qs_1.default.parse(qs_1.default.stringify(request.query));
        request.query = query;
        try {
            const page = +query.page > 0 ? +query.page : 1;
            const perPage = routeModule.perPage || DEFAULT_PER_PAGE;
            const offset = (page - 1) * perPage;
            const db = server.mongo.db(DB_NAME);
            request.collection = db.collection(routeModule.route.type);
            if (query.filter && typeof query.filter !== 'object') {
                throw new FnfError({
                    request,
                    reply,
                    status: 400,
                    title: 'Bad Request',
                    detail: `The filter query parameter must be an object for matching key/value pairs. Example: ?filter[attributes.title]=Something`,
                    code: 'BAD_REQUEST',
                });
            }
            if (query.order && query.order != 'ascending' && query.order != 'descending') {
                throw new FnfError({
                    request,
                    reply,
                    status: 400,
                    title: 'Bad Request',
                    detail: `The order query parameter must either be ascending or descending.`,
                    code: 'BAD_REQUEST',
                });
            }
            const filters = Object.keys(query.filter || {}).reduce((arr, key) => {
                if (+query.filter[key] == query.filter[key]) {
                    arr.push({
                        $or: [
                            {
                                [key]: {
                                    $eq: +query.filter[key],
                                },
                            },
                            {
                                [key]: {
                                    $eq: query.filter[key],
                                },
                            },
                        ],
                    });
                }
                else if (query.filter[key] === 'true') {
                    arr.push({
                        $or: [
                            {
                                [key]: {
                                    $eq: true,
                                },
                            },
                            {
                                [key]: {
                                    $eq: query.filter[key],
                                },
                            },
                        ],
                    });
                }
                else if (query.filter[key] === 'false') {
                    arr.push({
                        $or: [
                            {
                                [key]: {
                                    $eq: false,
                                },
                            },
                            {
                                [key]: {
                                    $eq: query.filter[key],
                                },
                            },
                        ],
                    });
                }
                else {
                    arr.push({
                        [key]: {
                            $eq: query.filter[key],
                        },
                    });
                }
                return arr;
            }, []);
            const filterQuery = {
                'attributes.deletedAt': {
                    $exists: false,
                },
            };
            if (filters.length > 0) {
                filterQuery.$and = filters;
            }
            let sort = [[query.orderBy || 'attributes.createdAt', query.orderBy ? query.order || 'ascending' : query.order || 'descending']];
            request.filterQuery = filterQuery;
            request.total = yield request.collection.countDocuments(filterQuery);
            request.documents = yield request.collection.find(filterQuery, {
                projection: { _id: 0 },
                sort,
                limit: perPage,
                skip: offset,
            }).toArray();
            // if middleware did not call next, do not perform default actions
            let nextCalled = yield preHooks(request, reply, routeModule);
            if (!nextCalled)
                return;
            if (routeModule.getAll) {
                nextCalled = false;
                yield routeModule.getAll.bind(server)(request, reply, () => {
                    nextCalled = true;
                });
            }
            if (!nextCalled)
                return;
            const protocol = request.raw.headers['x-forwarded-proto'] ? request.raw.headers['x-forwarded-proto'] : request.raw.socket['encrypted'] ? 'https' : 'http';
            const host = request.raw.headers['x-forwarded-host'] ? request.raw.headers['x-forwarded-host'] : request.raw.headers.host;
            const prefix = request.raw.headers['x-forwarded-prefix'] ? request.raw.headers['x-forwarded-prefix'] : '';
            const url = new URL(`${protocol}://${host}${prefix}${request.raw.url}`);
            url.searchParams.delete('page');
            const first = url.href;
            url.searchParams.set('page', `${Math.ceil(request.total / perPage)}`);
            const last = url.href;
            url.searchParams.set('page', `${page - 1}`);
            const prev = offset - perPage < 0 ? null : url.href;
            url.searchParams.set('page', `${page + 1}`);
            const next = offset + perPage > request.total ? null : url.href;
            const included = query.include ? yield fetchIncludes.bind(server)(request.documents, query.include) : undefined;
            reply.status(200).send(Object.assign(Object.assign({}, yield exports.jsonapiResponse(request.documents || [], included)), { links: {
                    first,
                    last,
                    prev,
                    next,
                }, meta: {
                    pages: Math.ceil(request.total / perPage),
                    total: request.total,
                } }));
        }
        catch (e) {
            yield handleError(request, reply, routeModule, e);
        }
    });
});
const defaultRoutePost = (routeModule) => __awaiter(void 0, void 0, void 0, function* () {
    return (request, reply) => __awaiter(void 0, void 0, void 0, function* () {
        try {
            const db = server.mongo.db(DB_NAME);
            request.collection = db.collection(routeModule.route.type);
            // if middleware did not call next, do not perform default actions
            let nextCalled = yield preHooks(request, reply, routeModule);
            if (!nextCalled)
                return;
            if (routeModule.post) {
                nextCalled = false;
                yield routeModule.post.bind(server)(request, reply, () => {
                    nextCalled = true;
                });
            }
            if (!nextCalled)
                return;
            // default behavior
            const now = new Date();
            const body = request.body;
            if (!body.data.attributes.createdAt) {
                body.data.attributes.createdAt = now;
            }
            if (!body.data.attributes.updatedAt) {
                body.data.attributes.updatedAt = null;
            }
            if (body.data.relationships) {
                for (const relationshipName of Object.keys(body.data.relationships)) {
                    const relationship = body.data.relationships[relationshipName];
                    const collection = db.collection(relationship.type);
                    const document = yield collection.findOne({
                        'attributes.deletedAt': {
                            $exists: false,
                        },
                        id: {
                            $eq: relationship.id,
                        },
                    });
                    if (!document) {
                        throw new FnfError({
                            request,
                            reply,
                            status: 400,
                            title: 'Bad Request',
                            detail: `Relationship of type '${relationship.type}' by id '${relationship.id}' does not exist.`,
                            code: 'BAD_REQUEST',
                        });
                    }
                }
            }
            if (body.data.id) {
                const document = yield request.collection.findOne({
                    'attributes.deletedAt': {
                        $exists: false,
                    },
                    id: {
                        $eq: body.data.id,
                    },
                });
                if (document) {
                    throw new FnfError({
                        request,
                        reply,
                        status: 400,
                        title: 'Bad Request',
                        detail: `A resource with the supplied id '${body.data.id}' already exists.`,
                        code: 'BAD_REQUEST',
                    });
                }
            }
            else {
                body.data = Object.assign({ id: uuid_1.v4() }, body.data);
            }
            const result = yield request.collection.insertOne(body.data);
            const document = yield request.collection.findOne({
                'attributes.deletedAt': {
                    $exists: false,
                },
                _id: {
                    $eq: result.insertedId,
                },
            });
            if (document)
                delete document['_id'];
            reply.status(200).send(yield exports.jsonapiResponse(document));
        }
        catch (e) {
            yield handleError(request, reply, routeModule, e);
        }
    });
});
const defaultRoutePatch = (routeModule) => __awaiter(void 0, void 0, void 0, function* () {
    return (request, reply) => __awaiter(void 0, void 0, void 0, function* () {
        const params = request.params;
        try {
            const db = server.mongo.db(DB_NAME);
            request.collection = db.collection(routeModule.route.type);
            request.document = yield request.collection.findOne({
                'attributes.deletedAt': {
                    $exists: false,
                },
                id: {
                    $eq: params.id
                }
            });
            if (request.document)
                delete request.document['_id'];
            // if middleware did not call next, do not perform default actions
            let nextCalled = yield preHooks(request, reply, routeModule);
            if (!nextCalled)
                return;
            if (routeModule.post) {
                nextCalled = false;
                yield routeModule.post.bind(server)(request, reply, () => {
                    nextCalled = true;
                });
            }
            if (!nextCalled)
                return;
            if (!request.document) {
                throw new FnfError({
                    request,
                    reply,
                    status: 404,
                    title: 'Not Found',
                    code: 'NOT_FOUND',
                    detail: 'The requested resource was not found.',
                });
            }
            // default behavior
            const now = new Date();
            const body = request.body;
            if (typeof body.data.id !== 'undefined' && body.data.id != params.id) {
                throw new FnfError({
                    request,
                    reply,
                    status: 400,
                    title: 'Bad Request',
                    code: 'BAD_REQUEST',
                    detail: 'The id of a resource cannot be changed.',
                });
            }
            if (!body.data.attributes.updatedAt) {
                body.data.attributes.updatedAt = now;
            }
            if (body.data.relationships) {
                for (const relationshipName of Object.keys(body.data.relationships)) {
                    const relationship = body.data.relationships[relationshipName];
                    const collection = db.collection(relationship.type);
                    const document = yield collection.findOne({
                        'attributes.deletedAt': {
                            $exists: false,
                        },
                        id: {
                            $eq: relationship.id,
                        },
                    });
                    if (!document) {
                        throw new FnfError({
                            request,
                            reply,
                            status: 400,
                            title: 'Bad Request',
                            detail: `Relationship of type '${relationship.type}' by id '${relationship.id}' does not exist.`,
                            code: 'BAD_REQUEST',
                        });
                    }
                }
            }
            const keys = [];
            const recursiveFindKeys = (ref, prevKey = '') => {
                Object.keys(ref).forEach((key) => {
                    const value = ref[key];
                    const keyNames = [prevKey, key].filter(v => v);
                    const currentKey = keyNames.join('.');
                    if (typeof value !== 'object' || value == null) {
                        return keys.push(currentKey);
                    }
                    return recursiveFindKeys(value, currentKey);
                });
            };
            recursiveFindKeys(body.data, '');
            const result = yield request.collection.updateOne({
                'attributes.deletedAt': {
                    $exists: false,
                },
                id: {
                    $eq: request.document['id']
                }
            }, {
                $set: keys.reduce((accumulator, key) => {
                    accumulator[key] = get_1.default(body.data, key);
                    return accumulator;
                }, {}),
            });
            const document = yield request.collection.findOne({
                'attributes.deletedAt': {
                    $exists: false,
                },
                id: {
                    $eq: request.document['id']
                }
            });
            if (document)
                delete document['_id'];
            reply.status(200).send(yield exports.jsonapiResponse(document));
        }
        catch (e) {
            yield handleError(request, reply, routeModule, e);
        }
    });
});
const defaultRouteDelete = (routeModule) => __awaiter(void 0, void 0, void 0, function* () {
    return (request, reply) => __awaiter(void 0, void 0, void 0, function* () {
        try {
            yield preHooks(request, reply, routeModule);
            // default behavior
            const { id } = request.params;
            const now = new Date();
            const db = server.mongo.db(DB_NAME);
            const collection = db.collection(routeModule.route.type);
            const result = yield collection.updateOne({
                'attributes.deletedAt': {
                    $exists: false,
                },
                id: {
                    $eq: id,
                },
            }, {
                $set: {
                    'attributes.deletedAt': now,
                }
            });
            if (!result.modifiedCount) {
                throw new FnfError({
                    request,
                    reply,
                    status: 404,
                    title: 'Not Found',
                    code: 'NOT_FOUND',
                    detail: 'The requested resource was not found.',
                });
            }
            reply.status(204).send();
        }
        catch (e) {
            yield handleError(request, reply, routeModule, e);
        }
    });
});
exports.jsonapiResponse = (data, included) => __awaiter(void 0, void 0, void 0, function* () {
    if (Array.isArray(data)) {
        return {
            data,
            included,
        };
    }
    else {
        return {
            data,
            included,
        };
    }
});
exports.jsonapiError = ({ request, reply, status, title, detail, code, }) => __awaiter(void 0, void 0, void 0, function* () {
    return {
        errors: [{
                id: `${process.pid}-${request.id}-${+new Date()}`,
                status,
                code,
                title,
                detail,
            }]
    };
});
//# sourceMappingURL=app.js.map